(**************************************************************************)
(*                                                                        *)
(*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       *)
(*  Kim Nguyen, Laurent Sartre                                            *)
(*                                                                        *)
(*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      *)
(*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 *)
(*                                                                        *)
(*  https://www.informatique-mpi.fr/                                      *)
(*                                                                        *)
(**************************************************************************)

(* recherche d'un motif dans un texte avec l'algorithme de Rabin--Karp *)

let b = 256 (* base du calcul *)
let p = 1_869_461_003 (* un nombre premier tel que `p*p` tient dans `int` *)

(* le hachage de `s[i..i+len[` *)
let hash s i len =
  assert (i + len <= String.length s);
  let rec compute h j =
    if j = i+len then h else compute ((h * b + Char.code s.[j]) mod p) (j+1) in
  compute 0 i

(* x^n modulo p *)
let rec power_mod x n =
  if n = 0 then
    1
  else
    let y = power_mod ((x * x) mod p) (n / 2) in
    if n mod 2 = 0 then y else (y * x) mod p

(* affiche toutes les occurrences de `m` dans `t` avec Rabin--Karp
   et renvoie le nombre d'occurrences

   note : si le motif `m` est vide, renvoie directement lt+1
   (sans afficher toutes les chaînes "occurrence ...") *)
let rabin_karp m t =
  let lm = String.length m and lt = String.length t in
  if lm = 0 then lt + 1 else
  if lt < lm then 0 else
  let count = ref 0 in
  let sh = power_mod b (lm - 1) in
  let hm = hash m 0 lm and ht = ref (hash t 0 lm) in
  for i = 0 to lt - lm do
    if !ht = hm && String.sub t i lm = m then (
      Printf.printf "occurrence à la position %d\n" i;
      incr count
    );
    if i < lt - lm then (
      ht := (!ht + p - (Char.code t.[i] * sh) mod p) mod p;
      ht := (!ht * b + Char.code t.[i + lm]) mod p
    )
  done;
  !count

(* Pour utilisation combinée avec LZW ; cf lzw.ml *)
module Buffer = struct

  type t = {
    m: string; (* le motif *)
    lm: int; (* sa longueur *)
    target: int; (* sa valeur de hachage *)
    mutable pos: int; (* nombre total de caractères déjà vus *)
    chars: Bytes.t;
    mutable current: int; (* le hachage des `lm` derniers caractères *)
    shift: int; (* b^(lm - 1) mod p *)
    mutable count: int; (* nombre total d'occurrences *)
  }

  let create m =
    let lm = String.length m in
    if lm = 0 then invalid_arg "create";
    { m; lm; target = hash m 0 lm;
      pos = 0; chars = Bytes.make lm '\000'; current = 0;
      shift = power_mod b (lm - 1); count = 0; }

  let matches t =
    let rec check i =
      i = t.lm ||
      Bytes.get t.chars ((t.pos + i) mod t.lm) = t.m.[i] && check (i+1) in
    check 0

  let add_char t c =
    let i = t.pos mod t.lm in
    let h = t.current in
    let h = (h + p - (Char.code (Bytes.get t.chars i) * t.shift) mod p) mod p in
       (* on a initialisé avec '\000' donc on n'a donc pas besoin
          de faire de cas particulier pour les lm premiers caractères *)
    let h = (h * b + Char.code c) mod p in
    t.current <- h;
    Bytes.set t.chars i c;
    t.pos <- t.pos + 1;
    if t.pos >= t.lm && t.current = t.target && matches t then (
      Printf.printf "occurrence à la position %d\n" (t.pos - t.lm);
      t.count <- t.count + 1
    )

  let add_string t s =
    String.iter (add_char t) s

  let count t =
    t.count

end

